// jthz backup

#include <vcl.h>
#pragma hdrstop
#include <iostream.h>
#include <inifiles.hpp>
#include <stdlib.h> // exports atexit

#define FOUND_SUCCESS 0

void cleanup();
int error(HWND,PCSTR);
bool DirectoryExists(String);
void ForceDirectories(String);
void tree_1(String,bool);
void tree_1Process(String,TSearchRec,bool);
void tree_2(String,bool);
void tree_2Process(String,TSearchRec,bool);
bool skip(String);

TStringList *recursed=0,*non_recursed=0,*exclude=0;
String destination;

void cleanup(){
    // atexit handler
    delete exclude;
    exclude=0;
    delete recursed;
    recursed=0;
    delete non_recursed;
    non_recursed=0;
}

int error(HWND hWnd,PCSTR s){
    MessageBox(hWnd,s,"SORRY",MB_ICONINFORMATION);
    return 0;
}

bool DirectoryExists(String Dir){
    // hz translation from Delphi/filectrl original
    // to avoid overhead of linking in vclx40.lib (247K)
    int Code=GetFileAttributes(Dir.c_str());
    return (Code!=-1)&&(FILE_ATTRIBUTE_DIRECTORY&Code);
}

void ForceDirectories(String Dir){
    // hz translation from Delphi/filectrl original
    // to avoid overhead of linking in vclx40.lib (247K)
    if(!Dir.Length())
        return;
    if(*AnsiLastChar(Dir)=='\\')
        Dir.Delete(Dir.Length(),1);
    if((Dir.Length()<3)||DirectoryExists(Dir)||(ExtractFilePath(Dir)==Dir))
        return;
    ForceDirectories(ExtractFilePath(Dir));
    CreateDir(Dir);
}

void tree_1(String inputDir,bool bRecurse){
    // recurses directory if bRecurse==true
    if(!inputDir.Length())
        return;
    if(*AnsiLastChar(inputDir)!='\\')
        inputDir+="\\";
    TSearchRec sr;
    if(FindFirst(inputDir+"*.*",faAnyFile,sr)==FOUND_SUCCESS){
        tree_1Process(inputDir,sr,bRecurse);
        while(FindNext(sr)==FOUND_SUCCESS)
            tree_1Process(inputDir,sr,bRecurse);
        FindClose(sr);
    }
}

void tree_1Process(String path,TSearchRec tsr,bool bRecurse){
    if(tsr.Attr&faDirectory){
        if(bRecurse&&tsr.Name[1]!='.')
            tree_1(path+tsr.Name,true);
    }
    else if(!skip(tsr.Name)){
        // build copy-destination (global 'destination' ends with \)
        String destFile=destination+path[1]+"\\"+
                path.SubString(4,path.Length()-3)+tsr.Name;
        TSearchRec srDest;
        if(FindFirst(destFile,faAnyFile,srDest)==FOUND_SUCCESS){
            // file exists; is our file newer?
            // use y2k-proof 64bit arithmetic     =8)~
            LARGE_INTEGER sourceTimeDate,destTimeDate;
            CopyMemory(&sourceTimeDate,&tsr.FindData.ftLastWriteTime,
                    sizeof(LARGE_INTEGER));
            CopyMemory(&destTimeDate,&srDest.FindData.ftLastWriteTime,
                    sizeof(LARGE_INTEGER));
            if(sourceTimeDate.QuadPart>destTimeDate.QuadPart){
                cout<<path.c_str()<<tsr.Name.c_str()<<endl;
                cout<<"File is updated - copying..."<<endl;
                // remove RHSA
                SetFileAttributes(destFile.c_str(),FILE_ATTRIBUTE_NORMAL);
                CopyFile((path+tsr.Name).c_str(),destFile.c_str(),FALSE);
            }
            FindClose(srDest);
        }
        else{
            cout<<path.c_str()<<tsr.Name.c_str()<<endl;
            cout<<"File is new - copying..."<<endl;
            ForceDirectories(ExtractFilePath(destFile));
            CopyFile((path+tsr.Name).c_str(),destFile.c_str(),FALSE);
        }
    }
}

void tree_2(String inputDir,bool bRecurse){
    // recurses directory if bRecurse==true
    if(!inputDir.Length())
        return;
    if(*AnsiLastChar(inputDir)!='\\')
        inputDir+="\\";
    TSearchRec sr;
    if(FindFirst(inputDir+"*.*",faAnyFile,sr)==FOUND_SUCCESS){
        tree_2Process(inputDir,sr,bRecurse);
        while(FindNext(sr)==FOUND_SUCCESS)
            tree_2Process(inputDir,sr,bRecurse);
        FindClose(sr);
    }
}

void tree_2Process(String path,TSearchRec tsr,bool bRecurse){
    if(tsr.Attr&faDirectory){
        if(bRecurse&&tsr.Name[1]!='.')
            tree_2(path+tsr.Name,true);
    }
    else{
        // map destination-path back to source-path
        String destFile=path+tsr.Name;
        int len=destination.Length();
        String checkFile=destFile.SubString(len,destFile.Length()-len+1);
        checkFile[1]=destFile[len+1];
        checkFile[2]=':';
        checkFile[3]='\\';
        if(!FileExists(checkFile)){
            cout<<destFile.c_str()<<endl;
            cout<<"Deleting obsolete file..."<<endl;
            // remove RHSA
            SetFileAttributes(destFile.c_str(),FILE_ATTRIBUTE_NORMAL);
            DeleteFile(destFile.c_str());
        }
    }
}

bool skip(String fileName){
    // returns true if fileName should be skipped
    // based on entries in global TStringList exclude
    // if name begins with a dot, it's considered an extension
    // else it's considered a complete filename (w/o the path)
    int f_len=fileName.Length();
    if(!f_len)
        return false;
    for(int i=0;i<exclude->Count;++i){
        int e_len=exclude->Strings[i].Length();
        if(!e_len)
            continue;
        if(exclude->Strings[i][1]=='.'){ // it's an extension
            if(f_len<e_len) // avoid out-of-bounds
                continue;
            if(!AnsiCompareText(fileName.SubString(f_len-e_len+1,e_len),
                    exclude->Strings[i]))
                return true; // skip me
        }
        else{ // it's a filename
            if(!AnsiCompareText(fileName,exclude->Strings[i]))
                return true; // skip me
        }
    }
    return false; // go ahead
}

int main(int argc,char* argv[]){
    char szTitle[256];
    HWND hConsole=NULL;
    // get our own handle
    if(GetConsoleTitle(szTitle,sizeof(szTitle)))
        hConsole=FindWindow(NULL,szTitle);
    // see if we are started with a cmdline parm
    if(argc!=2)
        return error(hConsole,"Run me with a parameter");
    // try to read info from file
    recursed=new TStringList(); // recurse these directories
    non_recursed=new TStringList(); // just scan these directories
    exclude=new TStringList(); // exclude these file extensions
    TIniFile *ini=new TIniFile(argv[1]);
    recursed->CommaText=ini->ReadString("jthz","recursed","");
    non_recursed->CommaText=ini->ReadString("jthz","non-recursed","");
    destination=ini->ReadString("jthz","destiny","");
    exclude->CommaText=ini->ReadString("jthz","xclude","");
    delete ini;
    // install cleaner-handler
    atexit(cleanup);
    // check validity of input
    if(!destination.Length())
        return error(hConsole,"No destination scanned");
    if(*AnsiLastChar(destination)!='\\')
        destination+="\\";
    // make sure destination exists
    ForceDirectories(destination);
    // now see if we have access to destination
    if(!DirectoryExists(destination))
        return error(hConsole,"Invalid destination");
    // source(s) valid?
    if(!recursed->Count&&!non_recursed->Count)
        return error(hConsole,"No source-directories scanned");
    for(int i=0;i<recursed->Count;++i)
        if(!DirectoryExists(recursed->Strings[i]))
            return error(hConsole,"Non-existing directory in recursed dirs");
    for(int i=0;i<non_recursed->Count;++i)
        if(!DirectoryExists(non_recursed->Strings[i]))
            return error(hConsole,"Non-existing directory in non-recursed entry");
    // first round; copy new files from source to destination
    cout<<"Hunting down new files..."<<endl<<endl;
    for(int i=0;i<recursed->Count;++i)
        tree_1(recursed->Strings[i],true);
    for(int i=0;i<non_recursed->Count;++i)
        tree_1(non_recursed->Strings[i],false);
    // second round; remove obsolete files
    cout<<"Hunting down obsolete files..."<<endl<<endl;
    for(int i=0;i<recursed->Count;++i){
        String tmp=recursed->Strings[i];
        if(!tmp.Length())
            continue;
        tree_2(destination+tmp[1]+"\\"+tmp.SubString(4,tmp.Length()-3),true);
    }
    for(int i=0;i<non_recursed->Count;++i){
        String tmp=non_recursed->Strings[i];
        if(!tmp.Length())
            continue;
        tree_2(destination+tmp[1]+"\\"+tmp.SubString(4,tmp.Length()-3),false);
    }
    return 0;
}
